use crate::Numeric;

pub use self::generic::{ColumnVector, GenericMatrix, RowVector, SquareMatrix};
pub use self::view::{View, ViewMut};

mod generic;
mod view;

/// Abstraction for all kinds of fixed size matrices.
pub trait Matrix<T: Numeric, const ROWS: usize, const COLUMNS: usize> {
    /// Get the value of the given cell.
    ///
    /// # Panics
    ///
    /// This function panics, if the row or column are outside of the valid range of the matrix.
    fn get(&self, row: usize, col: usize) -> T;
}

pub trait MatrixMut<T: Numeric, const ROWS: usize, const COLUMNS: usize>: Matrix<T, ROWS, COLUMNS> {
    /// Set the value of the given cell.
    ///
    /// # Panics
    ///
    /// This function panics, if the row or column are outsize of the valid range of the matrix.
    ///
    /// Additionally it may panic, if the value is invalid for the given cell.
    /// E.g. cells outsize of the main diagonal of a `DiagonalMatrix` can only be set to zero.
    fn set(&mut self, row: usize, col: usize, value: T);
}

/// Helper class representing the result of a matrix comparison.
#[derive(Eq, PartialEq, Debug, Copy, Clone)]
pub enum ComparisonResult {
    /// The two matrices are considered unequal.
    NotEqual,

    /// The two matrices are considered equal.
    Equal,
}

/// Compare two matrices for equality with a given tolerance.
/// This is used as the implementation by all the `PartialEq` implementations of the different matrix types.
pub fn compare<T: Numeric, const ROWS: usize, const COLUMNS: usize, RHS: Matrix<T, ROWS, COLUMNS>, LHS: Matrix<T, ROWS, COLUMNS>>(
    rhs: &RHS,
    lhs: &LHS,
    tolerance: T,
) -> ComparisonResult {
    for r in 0..ROWS {
        for c in 0..COLUMNS {
            if !rhs.get(r, c).is_equal_to(&lhs.get(r, c), tolerance) {
                return ComparisonResult::NotEqual;
            }
        }
    }
    ComparisonResult::Equal
}

pub fn add<
    T: Numeric, const ROWS: usize, const COLUMNS: usize,
    RESULT: MatrixMut<T, ROWS, COLUMNS>,
    LHS: Matrix<T, ROWS, COLUMNS>,
    RHS: Matrix<T, ROWS, COLUMNS>
>(result: &mut RESULT, lhs: &LHS, rhs: &RHS) {
    for r in 0..ROWS {
        for c in 0..COLUMNS {
            result.set(r, c, lhs.get(r, c) + rhs.get(r, c));
        }
    }
}

pub fn subtract<
    T: Numeric, const ROWS: usize, const COLUMNS: usize,
    RESULT: MatrixMut<T, ROWS, COLUMNS>,
    LHS: Matrix<T, ROWS, COLUMNS>,
    RHS: Matrix<T, ROWS, COLUMNS>
>(result: &mut RESULT, lhs: &LHS, rhs: &RHS) {
    for r in 0..ROWS {
        for c in 0..COLUMNS {
            result.set(r, c, lhs.get(r, c) - rhs.get(r, c));
        }
    }
}

pub fn multiply<
    T: Numeric, const ROWS: usize, const COMMON: usize, const COLUMNS: usize,
    RESULT: MatrixMut<T, ROWS, COLUMNS>,
    LHS: Matrix<T, ROWS, COMMON>,
    RHS: Matrix<T, COMMON, COLUMNS>
>(result: &mut RESULT, lhs: &LHS, rhs: &RHS) {
    for i in 0..ROWS {
        for j in 0..COLUMNS {
            for k in 0..COMMON {
                result.set(i, j, result.get(i, j) + lhs.get(i, k) * rhs.get(k, j));
            }
        }
    }
}
