use std::fmt::{Debug, Formatter};
use std::marker::PhantomData;
use crate::{GenericMatrix, Matrix, MatrixMut, Numeric};
use crate::types::matrix::{compare, ComparisonResult};

pub struct View<
    'a,
    T: Numeric,
    const ROWS: usize,
    const COLUMNS: usize,
    const BASE_ROWS: usize,
    const BASE_COLUMNS: usize,
    M: Matrix<T, BASE_ROWS, BASE_COLUMNS>
> {
    base: &'a M,
    origin: (usize, usize),
    phantom: PhantomData<*const T>,
}

pub struct ViewMut<
    'a,
    T: Numeric,
    const ROWS: usize,
    const COLUMNS: usize,
    const BASE_ROWS: usize,
    const BASE_COLUMNS: usize,
    M: MatrixMut<T, BASE_ROWS, BASE_COLUMNS>
> {
    base: &'a mut M,
    origin: (usize, usize),
    phantom: PhantomData<*const T>,
}

impl<
    'a,
    T: Numeric,
    const ROWS: usize,
    const COLUMNS: usize,
    const BASE_ROWS: usize,
    const BASE_COLUMNS: usize,
    M: Matrix<T, BASE_ROWS, BASE_COLUMNS>
> Debug for View<'a, T, ROWS, COLUMNS, BASE_ROWS, BASE_COLUMNS, M> {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        let mut mat = GenericMatrix::<T, ROWS, COLUMNS>::new();
        for r in 0..ROWS {
            for c in 0..COLUMNS {
                mat.data[r][c] = self.get(r, c);
            }
        }

        f
            .debug_struct("View")
            .field("data", &mat.data)
            .finish()
    }
}

impl<
    'a,
    T: Numeric,
    const ROWS: usize,
    const COLUMNS: usize,
    const BASE_ROWS: usize,
    const BASE_COLUMNS: usize,
    M: MatrixMut<T, BASE_ROWS, BASE_COLUMNS>
> Debug for ViewMut<'a, T, ROWS, COLUMNS, BASE_ROWS, BASE_COLUMNS, M> {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        let mut mat = GenericMatrix::<T, ROWS, COLUMNS>::new();
        for r in 0..ROWS {
            for c in 0..COLUMNS {
                mat.data[r][c] = self.get(r, c);
            }
        }

        f
            .debug_struct("ViewMut")
            .field("data", &mat.data)
            .finish()
    }
}

impl<
    'a,
    T: Numeric,
    const ROWS: usize,
    const COLUMNS: usize,
    const BASE_ROWS: usize,
    const BASE_COLUMNS: usize,
    M: Matrix<T, BASE_ROWS, BASE_COLUMNS>
> View<'a, T, ROWS, COLUMNS, BASE_ROWS, BASE_COLUMNS, M> {
    pub fn new(base: &'a M, origin: (usize, usize)) -> Self {
        Self {
            base,
            origin,
            phantom: PhantomData,
        }
    }
}

impl<
    'a,
    T: Numeric,
    const ROWS: usize,
    const COLUMNS: usize,
    const BASE_ROWS: usize,
    const BASE_COLUMNS: usize,
    M: MatrixMut<T, BASE_ROWS, BASE_COLUMNS>
> ViewMut<'a, T, ROWS, COLUMNS, BASE_ROWS, BASE_COLUMNS, M> {
    pub fn new(base: &'a mut M, origin: (usize, usize)) -> Self {
        Self {
            base,
            origin,
            phantom: PhantomData,
        }
    }
}

impl<
    'a,
    T: Numeric,
    const ROWS: usize,
    const COLUMNS: usize,
    const BASE_ROWS: usize,
    const BASE_COLUMNS: usize,
    M: Matrix<T, BASE_ROWS, BASE_COLUMNS>
> Matrix<T, ROWS, COLUMNS> for View<'a, T, ROWS, COLUMNS, BASE_ROWS, BASE_COLUMNS, M> {
    fn get(&self, row: usize, col: usize) -> T {
        let (base_row, base_col) = self.origin;
        self.base.get(base_row + row, base_col + col)
    }
}

impl<
    'a,
    T: Numeric,
    const ROWS: usize,
    const COLUMNS: usize,
    const BASE_ROWS: usize,
    const BASE_COLUMNS: usize,
    M: MatrixMut<T, BASE_ROWS, BASE_COLUMNS>
> Matrix<T, ROWS, COLUMNS> for ViewMut<'a, T, ROWS, COLUMNS, BASE_ROWS, BASE_COLUMNS, M> {
    fn get(&self, row: usize, col: usize) -> T {
        let (base_row, base_col) = self.origin;
        self.base.get(base_row + row, base_col + col)
    }
}

impl<
    'a,
    T: Numeric,
    const ROWS: usize,
    const COLUMNS: usize,
    const BASE_ROWS: usize,
    const BASE_COLUMNS: usize,
    M: MatrixMut<T, BASE_ROWS, BASE_COLUMNS>
> MatrixMut<T, ROWS, COLUMNS> for ViewMut<'a, T, ROWS, COLUMNS, BASE_ROWS, BASE_COLUMNS, M> {
    fn set(&mut self, row: usize, col: usize, value: T) {
        let (base_row, base_col) = self.origin;
        self.base.set(base_row + row, base_col + col, value);
    }
}

impl<
    'a,
    T: Numeric,
    const ROWS: usize,
    const COLUMNS: usize,
    const BASE_ROWS: usize,
    const BASE_COLUMNS: usize,
    M: Matrix<T, BASE_ROWS, BASE_COLUMNS>,
    RHS: Matrix<T, ROWS, COLUMNS>
> PartialEq<RHS> for View<'a, T, ROWS, COLUMNS, BASE_ROWS, BASE_COLUMNS, M> {
    fn eq(&self, other: &RHS) -> bool {
        compare(self, other, T::epsilon()) == ComparisonResult::Equal
    }

    fn ne(&self, other: &RHS) -> bool {
        compare(self, other, T::epsilon()) == ComparisonResult::NotEqual
    }
}

impl<
    'a,
    T: Numeric,
    const ROWS: usize,
    const COLUMNS: usize,
    const BASE_ROWS: usize,
    const BASE_COLUMNS: usize,
    M: MatrixMut<T, BASE_ROWS, BASE_COLUMNS>,
    RHS: Matrix<T, ROWS, COLUMNS>
> PartialEq<RHS> for ViewMut<'a, T, ROWS, COLUMNS, BASE_ROWS, BASE_COLUMNS, M> {
    fn eq(&self, other: &RHS) -> bool {
        compare(self, other, T::epsilon()) == ComparisonResult::Equal
    }

    fn ne(&self, other: &RHS) -> bool {
        compare(self, other, T::epsilon()) == ComparisonResult::NotEqual
    }
}

#[cfg(test)]
mod tests {
    use crate::{ColumnVector, RowVector, SquareMatrix};
    use crate::types::matrix::multiply;

    #[test]
    fn view_of_generic_matrix() {
        let mat = SquareMatrix::diagonal(&[1, 2, 3, 4]);
        let view = mat.view::<2, 2>(1, 1);
        assert_eq!(view, SquareMatrix::diagonal(&[2, 3]));
    }

    #[test]
    fn multiply_views_inside_generic_matrix() {
        let mut mat = SquareMatrix::from([
            [0, 4, -2, -1],
            [1, 0, 0, 0],
            [3, 0, 0, 0],
            [-5, 0, 0, 0],
        ]);

        let a = ColumnVector::from(mat.view::<3, 1>(1, 0));
        let b = RowVector::from(mat.view::<1, 3>(0, 1));
        multiply(&mut mat.view_mut::<3, 3>(1, 1), &a, &b);

        assert_eq!(
            mat,
            SquareMatrix::from([
                [0, 4, -2, -1],
                [1, 4, -2, -1],
                [3, 12, -6, -3],
                [-5, -20, 10, 5],
            ])
        );
    }
}
